/*
 * Copyright (C) 2019 Tristan Muller <tristan.muller@cirad.fr>.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA 02110-1301  USA
 */
package fr.amap.viewer3d;

import com.jogamp.nativewindow.util.Point;
import com.jogamp.opengl.GLCapabilities;
import com.jogamp.opengl.GLProfile;
import com.jogamp.opengl.awt.GLCanvas;
import com.jogamp.opengl.util.FPSAnimator;
import fr.amap.viewer3d.event.BasicEvent;
import fr.amap.viewer3d.event.EventManager;
import fr.amap.viewer3d.input.InputKeyListener;
import fr.amap.viewer3d.input.InputMouseAdapter;
import fr.amap.viewer3d.renderer.JoglListener;
import fr.amap.viewer3d.renderer.MinimalKeyAdapter;
import fr.amap.viewer3d.renderer.MinimalMouseAdapter;
import java.awt.Color;
import java.awt.event.KeyListener;
import java.awt.event.MouseListener;
import javax.swing.JPanel;

/**
 * Class to implement a 3D panel to a Swing/awt/javaFX element.
 * To implement it in a JavaFX context, you need to use a "SwingNode".
 * 
 * @author Tristan Muller <tristan.muller@cirad.fr>
 */
public final class Viewer3D extends JPanel {
    /**
     * Object used to create a rendering loop.
     */
    private final FPSAnimator animator;
    
    /**
     * OpenGL context where all drawing functions will be called.
     */
    private final JoglListener joglContext;
    
    /**
     * The OpenGL canvas where all 3D drawing will be displayed.
     */
    private final GLCanvas canvas;
    
    /**
     * The Event manager of this element and his children.
     */
//    private final EventManager minimalEventMgr;
    
    /**
     * If dynamic drawing process is enabled (by default not).
     */
    private boolean dynamicDraw = false;
    
    /**
     * Default constructor.
     * Initialize all elements to create a displayable panel.
     * (OpenGL, Animator, Events listeners ...)
     * @param posX The position of this element on the X axis (in px).
     * @param posY The position of this element on the Y axis (in px).
     * @param width The width of this element (in px).
     * @param height The height of this element (in px).
     * @param title The title of this element.
     */
    public Viewer3D(int posX, int posY, int width, int height, String title) {
        super();
        
        /* Get the OpenGL profile */
        GLProfile glp = GLProfile.getMaximum(false);
        GLCapabilities caps = new GLCapabilities(glp);
        caps.setDoubleBuffered(true);
        
        /* Init the animator to display at each 60 frames per seconds */
        this.animator = new FPSAnimator(60);
        
        /* Create a new GLCanvas which will draw all opengl functions */
        this.canvas = new GLCanvas(caps);
        this.canvas.setSize(width, height);
        
        /* Bind the canvas and the animator to create a rendering loop */
        this.animator.add(canvas);
        
        canvas.setBackground(Color.red);
        
        /* Init the JOGL listener */
        this.joglContext = new JoglListener(this.animator);
        
        /* Init the size and the location */
        this.rescale(posX, posY, width, height);
        
        this.add(canvas);
        
        /* Link the OpenGL's context to the OpenGL's canvas */
        this.canvas.addGLEventListener(this.joglContext);
        
        /* Add some input listeners */
        InputKeyListener inputKeyListener = new InputKeyListener();
        InputMouseAdapter inputMouseAdapter = new InputMouseAdapter();
        
        /* Init an event manager */
        EventManager minimalEventMgr = new EventManager(inputMouseAdapter, inputKeyListener) {
            @Override
            public void updateEvents() {
                
                if(!animator.isPaused() && !joglContext.isDynamicDraw() &&
                        !mouse.isButtonDown(InputMouseAdapter.Button.LEFT) && 
                        !mouse.isButtonDown(InputMouseAdapter.Button.RIGHT)){
                    //this function cost time, it should not be called at each updateEvents method call
                    animator.pause();
                }
            }
        };
        
        /* Add the previously created event manager */
//        addEventListener(minimalEventMgr);
        
        /* Create a basic events manager and add it to this element */
        BasicEvent basicEvents = new BasicEvent(joglContext, inputMouseAdapter, inputKeyListener);
//        addEventListener(basicEvents);
        
        /* basic input adapters for waking up animator if necessary */
        MinimalKeyAdapter minimalKeyAdapter = new MinimalKeyAdapter(animator);
//        canvas.addKeyListener((KeyListener) minimalKeyAdapter);

        MinimalMouseAdapter minimalMouseAdapter = new MinimalMouseAdapter(animator, dynamicDraw);
//        canvas.addMouseListener((MouseListener) minimalMouseAdapter);

        canvas.setVisible(true);
        animator.start();
    }
    
    /**
     * Change the location and the size of this element in the screen space.
     * @param posX The new position of this element on the X axis (in px).
     * @param posY The new position of this element on the Y axis (in px).
     * @param width The new width of this element (in px).
     * @param height The new height of this element (in px).
     */
    public void rescale(int posX, int posY, int width, int height) {
        this.setLocation(posX, posY);
        this.setSize(width, height);
        
        //this.canvas.setLocation(posX, posY);
        this.canvas.setSize(width, height);
                
        if (this.joglContext != null) {
            joglContext.getScene().setWidth(width);
            joglContext.getScene().setHeight(height);
        }
    }
    
    /**
     * Add key listener and mouse lister to this canvas and add event manager 
     * to this JOGL listener.
     * @param eventManager 
     */
    public final void addEventListener(EventManager eventManager){
        this.canvas.addKeyListener((KeyListener) eventManager.getKeyboard());
        this.canvas.addMouseListener((MouseListener) eventManager.getMouse());
        this.joglContext.addEventListener(eventManager);
    }

    public void setDynamicDraw(boolean dynamicDraw) {
        this.dynamicDraw = dynamicDraw;
    }
    
    public fr.amap.viewer3d.object.scene.Scene getScene(){
        return getJoglContext().getScene();
    }
    
    public JoglListener getJoglContext() {
        return joglContext;
    }
    
    public GLCanvas getRenderFrame() {
        return canvas;
    }
    
    public Point getPosition(){
        try{
            java.awt.Point locationOnScreen = this.getLocationOnScreen();
            return new Point((int) locationOnScreen.getX(), (int) locationOnScreen.getY());
        }catch(RuntimeException ex){
            return new Point(0, 0);
        }
    }
    
    public void close(){
        if(canvas.isVisible()){
            canvas.setVisible(false);
            canvas.destroy();
        }
    }
}
